#include"diagnostics.h"
#include"IO.h"
#include"cellmanagement.h"
#include"forcelaw.h"
#define PI 3.1415926536
#define TWO16 1.122462048
// Global variables
extern cell supercell;
extern IO readwrite;
extern forcelaw forces;
int nmobile, numclusters, numimmclusters, ninstrings, numstrings;
vector <int> pindices, mobileparticleids, mpidscopy, sizecounts, stringlengthcounts;
vector <int> immobileparticleids, immpidscopy, immpsizecounts;
vector <bool> ismobile, isimmobile, ispartofstring;
vector<double> particledisps, maxparticledisps, RgSqVsSize, immpRgSqVsSize, StringRgSqVsLength;

// Functions called within this file
void idmobileparticles(); // Identify the most-mobile particles
void findmobileparticleclusters(); // Analyze their clustering
void idimmobileparticles(); // Identify the least-mobile particles
void findimmobileparticleclusters(); // Analyze their clustering
void findstrings();  // Identify the mobile-particle strings

double adjcutoff = 1.25;  // clusters and strings consist of particles with distances < cutoff*sigma_ij
// 1.25 is a value typically used for d = 3; see e.g. Starr/Sastry/Douglas JCP 2013

// Calculate thermodynamic quantities
// Uses new generalized approach to evaluate forces, pair energies
void diagnostics::getthermo(bool outputflag) {
    double e, ep, ek, t, p, z, msdincr, namsdincr, m4d;  // per-atom versions of E, EP, EK, etc.
    int i, j, k, u;
    double delc[d], delc0[d], delsq, del;
    double ft, fovr;
    double f[d];
    int totneigh;
    double sigmaij, sigmaijsq, epsilonij; // for polydisp force-mixing
    double facij, shift; // facij = (sigmaij/del)^n

// initialize thermo quantities
    if (step < nmsdstart) elapstime = step*dt;
    else elapstime = (step-thiswindowstart)*dt;
    EK = 0;     T = 0;      EP = 0;     P = 0;      Z = 0;      Znoratt = 0;
    msd = 0;    namsd = 0;      m4d = 0;    maxdisp = 0;
    f_ov = 0;  // overlap function (Ninarello PRX 2017 Eq. 16)
// For tracking the dynamics of the two components in a bidisperse system separately
    msd1 = 0;   m4d1 = 0;   ngp1 = 0;   fov1 = 0;
    msd2 = 0;   m4d2 = 0;   ngp2 = 0;   fov2 = 0;
// Similarly for the mobile and immobile particles
    msdm = 0;   m4dm = 0;   ngpm = 0;   fovm = 0;
    msdimm = 0;   m4dimm = 0;   ngpimm = 0;   fovimm = 0;
// Cutoff for the overlap function; see below
    double delcutoff;
    if (model == 1 || model == 2) delcutoff = .5;
    else if (model == 3) delcutoff = .5*TWO16;
    
// For jamming-related analyses
    for (i = 0; i < N; i++) Z_array[i] = 0;
    rattlers = 0;      floaters = 0; 
    double maxsqdisp = 0;  // maxsqdisp is max (nonaffine displacement)^2 of any atom
    
// Reset maxparticle disps so can correctly identify mobile particles over multiple time windows
    if (step == 0 || step == thiswindowstart) for (i = 0; i < N; i++) maxparticledisps[i] = 0;
    
#pragma omp parallel
{
    #pragma vector aligned
#pragma omp for schedule(guided) private(j,k,u,ek,t,sigmaij,sigmaijsq,epsilonij,delc,delc0,delsq,del,facij,ep,ft,fovr,f,p) reduction(+:EK,T,Z,EP,P,msd,namsd,m4d,f_ov,msd1,msd2,m4d1,m4d2,ngp1,ngp2,fov1,fov2)
    for (i = 0; i < N; i++) {
        ek = 0;
        for (k = 0; k < d; k++) ek += mass[i]*RVF[3*d*i+d+k]*RVF[3*d*i+d+k]/2;
        t = (2*ek)/d;
        EK += ek/N;       T += t/N;
// Set force on particle i to zero
        for (k = 0; k < d; k++) f[k] = 0;
// declaring delc0[k] here increases efficiency
        for (k = 0; k < d; k++) delc0[k] = -RVF[3*d*i+k];
// Loop over particle i's neighbors
// NOTE INCREMENTS FOR Z, EP, P, ARE DOUBLED BECAUSE CODE APPLIES NEWTON'S THIRD LAW
        for (j = 0; j < numneigh[i]; j++) {
// Apply mixing rules for particle diameters, force/energy prefactors
            u = allneigh[i*maxneigh + j];
            sigmaij = forces.getsigmaij(i,u);
            epsilonij = forces.getepsilonij(i,u);
            sigmaijsq = sigmaij*sigmaij;
            delsq = 0;
            for (k = 0; k < d; k++) {
                delc[k] = delc0[k] + RVF[3*d*u+k];
                if (delc[k] < -D/2.0) delc[k] += D;    else if (delc[k] > D/2.0) delc[k] -= D;
                delsq += delc[k]*delc[k];
            }
            if (delsq <= cutsq*sigmaijsq) {
                Z_array[i] += 1;
                Z_array[u] += 1;
                Z += 2.0/N;
// New generalized approach to force/energy calculation
                fovr = forces.getfovr(delsq,sigmaij,sigmaijsq,epsilonij);
                ep = forces.getepair(delsq,sigmaij,sigmaijsq,epsilonij);
                EP += 2*ep/(2*N);  // Should these factors of 2 be elimiinated?
                for (k = 0; k < d; k++) f[k] = -delc[k]*fovr;
// Here calculate pair contribution to P using virial formula.
                p = delsq*fovr;
                P += p/(d*vol);
            }
        }
// Calculate MSD (total + its nonaffine component)
        double dispthis[d], nadispthis[d];
        double sqdisp = 0;
        for (k = 0; k < d; k++) {
            dispthis[k] = (RVF[3*d*i + k] + image[d*i + k]*D - Rinit[d*i + k] - initimg[d*i + k]*Dinit);
            nadispthis[k] = (RVF[3*d*i + k] + D*image[d*i + k] - (D/Dinit)*Rinit[d*i + k] - Dinit*initimg[d*i + k]);
            namsd += nadispthis[k]*nadispthis[k]/N;
            sqdisp += nadispthis[k]*nadispthis[k];
        }
        msd += sqdisp/N;
        m4d += sqdisp*sqdisp/N;

        del = sqrt(sqdisp);
// f_ov as defined in Ninarello et al. Physical Review X 2017 (Eq. 16), modified by Adhikari et. al. J. Phys. Chem. B 2021
//        if (del < 0.1*d) f_ov += 1.0/N;
// Alt. definition (cutoff = 2^-5/6 particle diameter, better suited for comparison to stringlike-motion analyses since is same criterion used by Starr/Douglas 2013
        if (del < delcutoff*sigma[i]) f_ov += 1.0/N;
// For tracking the dynamics of large and small particles separately
        if (i < Nlarge) {
            msd1 += sqdisp/Nlarge;   m4d1 += sqdisp*sqdisp/Nlarge;
//            if (del < 0.1*d) fov1 += 1.0/Nlarge;
            if (del < delcutoff*sigma[i]) fov1 += 1.0/Nlarge;
        }
        else {
            msd2 += sqdisp/Nsmall;   m4d2 += sqdisp*sqdisp/Nsmall;
//            if (del < 0.1*d) fov2 += 1.0/Nsmall;
            if (del < delcutoff*sigma[i]) fov2 += 1.0/Nsmall;
        }
// Record displacements for mobile-particle analyses
        if (i < N) {
            particledisps[i] = del;
            if (del > maxparticledisps[i]) maxparticledisps[i] = del;
        }
    } // End of loop over the N particles
# pragma omp barrier
} // End of the parallel region

// If desired, identify mobile particles, dump their coordinates if appropriate
    if (mobileflag && outputflag) {
        idmobileparticles();
        if (step > nmsdstart && !((step-nmsdstart)%nmobdump)) readwrite.writemobilecoords(step);
        findmobileparticleclusters();
        for (i = 0; i < N; i++) ismobile[i] = false;
        for (i = 0; i < nmobile; i++)   ismobile[mobileparticleids[i]] = true;
        findstrings();  // Find the strings, analyze their statistical-geometrical properties
// Now calculate msd, m4d, f_ov for mobile particles
        for (i = 0; i < nmobile; i++) {
            del = particledisps[mobileparticleids[i]];
            msdm += del*del/nmobile;
            m4dm += del*del*del*del/nmobile;
            if (del < delcutoff*sigma[mobileparticleids[i]]) fovm += 1.0/nmobile;
//            if (del < .1*d) fovm += 1.0/nmobile;
        }
// Now do similarly for the immobile particles
        idimmobileparticles();
        findimmobileparticleclusters();
        for (i = 0; i < N; i++) isimmobile[i] = false;
        for (i = 0; i < nmobile; i++)   isimmobile[immobileparticleids[i]] = true;
        for (i = 0; i < nmobile; i++) {
            del = particledisps[immobileparticleids[i]];
            msdimm += del*del/nmobile;
            m4dimm += del*del*del*del/nmobile;
            if (del < delcutoff*sigma[immobileparticleids[i]]) fovimm += 1.0/nmobile;
//            if (del < .1*d) fovimm += 1.0/nmobile;
        }
    }

// Calculate non-Gaussian param & maximum displacement of large, small, mobile, immobile particles
    ngp = ((1.0*d)/(d+2))*m4d/(msd*msd) - 1.0;
    if (std::isnan(ngp)) ngp = 0;
    ngp1 = ((1.0*d)/(d+2))*m4d1/(msd1*msd1) - 1.0;
    if (std::isnan(ngp1)) ngp1 = 0;
    ngp2 = ((1.0*d)/(d+2))*m4d2/(msd2*msd2) - 1.0;
    if (std::isnan(ngp2)) ngp2 = 0;
    ngpm = ((1.0*d)/(d+2))*m4dm/(msdm*msdm) - 1.0;
    if (std::isnan(ngpm)) ngpm = 0;
    ngpm = ((1.0*d)/(d+2))*m4dm/(msdm*msdm) - 1.0;
    if (std::isnan(ngpm)) ngpm = 0;
    ngpimm = ((1.0*d)/(d+2))*m4dimm/(msdimm*msdimm) - 1.0;
    if (std::isnan(ngpimm)) ngpimm = 0;

// Calculate S(q,t) if desired
    if (sofqflag && outputflag) getsisf();
    
// Calculate <Z>, identify rattlers, floaters
// Rattlers are atoms with <= d contacts, floaters are atoms with zero contacts.
    for (i = 0; i < N; i++) if (Z_array[i] <= d) {
        rattlers++;
        if (Z_array[i] == 0) floaters++;
    }
// Znoratt =  average coordination number of non-rattlers
    for (i = 0; i < N; i++) if (Z_array[i] > d) Znoratt += Z_array[i]/(1.0*(N - rattlers));

// Add the kinetic contribution to pressure and energy
    P += rho*T;
    E = EK + EP;
// For outputting phi during NPT runs
    vol = pow(D,d);
    phi = totpartvol/vol;
    rho = N/vol;
    
// Output thermodynamics summary
    if (outputflag) {
        cout << step << " " << elapstime << " " << T << " " << rho << " " << phi << " " << P << " " << Z << " " << EP << " " << msd << " " << ngp << " " << f_ov << " " << SISF << " " << avgmobileclustersize << " " << avgimmobileclustersize << " " << stringfrac << " " << avgstringsize << " " << swapsucc << " " << swapatt << " ";
        readwrite.outputthermo(step);
        readwrite.outputbidispdyn(step);
    }
    
    return;
}



// Calculate the van Hove correlation function
void diagnostics::getvanHove() {
    int i, k, bin;
    double vhc[vhcap], vhcsmall[vhcap], vhclarge[vhcap], vhcmobile[vhcap], vhcimmobile[vhcap];
    double dispthis[d], dispsq, disp;  // displacement of atom i
    for (i = 0; i < vhcap; i++) {  vhc[i] = 0;  vhcsmall[i] = 0;  vhclarge[i] = 0;   vhcmobile[i] = 0;   vhcimmobile[i] = 0;    }
    
    #pragma omp parallel for schedule(static,nchunk) private(dispthis,dispsq,k,disp,bin) reduction(+:vhc,vhcsmall,vhclarge)
    for (i = 0; i < N; i++) {
        dispsq = 0;
        for (k = 0; k < d; k++) {
            dispthis[k] = (RVF[3*d*i + k] + image[d*i + k]*D - (D/Dinit)*Rinit[d*i + k] - D*initimg[d*i + k]);
            dispsq += dispthis[k]*dispthis[k];
        }
        disp = sqrt(dispsq);
        bin = round(disp/.01);  // note the use of C++'s "round" function
        if (bin < vhcap) {
            vhc[bin] += 1.0/N;
            if (i < Nlarge) vhclarge[bin] += 1.0/Nlarge;
            else vhcsmall[bin] += 1.0/Nsmall;
            if (mobileflag && ismobile[i]) vhcmobile[bin] += 1.0/nmobile;
            else if (mobileflag && isimmobile[i]) vhcimmobile[bin] += 1.0/nmobile;
        }
    }
    readwrite.outputVHC(vhc,vhcsmall,vhclarge,vhcmobile,vhcimmobile,step);
    return;
}



// Calculate the pair correlation function
// This O(N) implementation requires rebuilding the atom lists with the minimum # of cells necessary to calculate g(r) out to r = paircorrcap
void diagnostics::getgofr() {
    int ncellold, ncelltmp, nbstmp;
    nbstmp = nbs;
    nbs = 3;
    ncellold = ncell;
    ncell = floor(D/paircorrcap);
    if (ncell < 2) { cout << "paircorrcap = " << paircorrcap << " is too high for the current D = " << D << ". g(r) calculation will be inaccurate." << endl;    exit(1);    }
    if (ncell < 3) ncell = 3;
// Rebuild the cell, perform the O(N) g(r) calculation
    supercell.setupsupercell();
    supercell.calcgofr();
// Restore the old cell and neighbor lists
    ncelltmp = ncell;
    ncell = ncellold;
    nbs = nbstmp;
    supercell.setupsupercell();
    supercell.buildneighborlist();
    neighlistbuilds++;

    return;
}
    

// Calculate the self-intermediate scattering function
// need specify q_min, q_max
void diagnostics::getsisf() {
    double q, qmin, qmax;
    qmin = 7.5;     qmax = 7.75;  // reasonable values for d = 3 2:1 KA model at T = .4, P = 0
    int n_q, nmin, nmax;
    nmin = floor(D*qmin/PI);    nmax = ceil(D*qmax/PI);
    double S_re, S_im;
    S_re = 0;   S_im = 0;
    double sre1, sim1, sre2, sim2;  // modeled after fov1, fov2 in getthermo()
    sre1 = 0;   sim1 = 0;   sre2 = 0;   sim2 = 0;
    double srem, simm;   // analogous quantities for mobile particles
    srem = 0;   simm = 0;
    double sreimm, simimm;   // analogous quantities for mobile particles
    sreimm = 0;   simimm = 0;
    int i, k, numq, numq1, numq2, numqm, numqimm;
    numq = 0;   numq1 = 0;  numq2 = 0;  numqm = 0;  numqimm = 0;
    double dispthis[d];
    #pragma omp parallel for schedule(static,nchunk) private(k,n,q,dispthis) reduction(+:S_re,S_im,numq,sre1,sim1,numq1,sre2,sim2,numq2,srem,simm,numqm,sreimm,simimm,numqimm)
    for (i = 0; i < N; i++) {
        for (k = 0; k < d; k++) dispthis[k] = (RVF[3*d*i + k] + image[d*i + k]*D - (D/Dinit)*Rinit[d*i + k] - Dinit*initimg[d*i + k]);
        for (n_q = nmin; n_q <= nmax; n_q++) {
            q = n_q*PI/D;
            for (k = 0; k < d; k++) {
                S_re += cos(q*dispthis[k]);     S_im += sin(q*dispthis[k]);     numq++;
                if (i < Nlarge) {   sre1 += cos(q*dispthis[k]);     sim1 += sin(q*dispthis[k]);     numq1++;   }
                else {   sre2 += cos(q*dispthis[k]);     sim2 += sin(q*dispthis[k]);     numq2++;   }
                if (ismobile[i]) {
                    srem += cos(q*dispthis[k]);     simm += sin(q*dispthis[k]);     numqm++;   }
                else if (isimmobile[i]) {
                    sreimm += cos(q*dispthis[k]);     simimm += sin(q*dispthis[k]);     numqimm++;   }
            }
        }
    }
    S_re /= numq;       S_im /= numq;
    SISF = S_re*S_re + S_im*S_im;
    sre1 /= numq1;      sim1 /= numq1;
    sisf1 = sre1*sre1 + sim1*sim1;
    sre2 /= numq2;      sim2 /= numq2;
    sisf2 = sre2*sre2 + sim2*sim2;
    srem /= numqm;      simm /= numqm;
    sisfm = srem*srem + simm*simm;
    sreimm /= numqimm;      simimm /= numqimm;
    sisfimm = sreimm*sreimm + simimm*simimm;
    
    return;
}



// Set up the mobile-particle-ID machinery
void diagnostics::initmobile() {
    double mobilefrac = .065;  // fraction of particles classified as "mobile": need adjust a la Douglas' shear-band JCP to optimize mobile-particle-cluster and string identification
    particledisps.resize(N);
    maxparticledisps.resize(N);
    ismobile.resize(N);     isimmobile.resize(N);
    for (int i = 0; i < N; i++) {   ismobile[i] = false;    isimmobile[i] = false;  }
    if (mobileflag) {
        pindices.resize(N);
        nmobile = ceil(mobilefrac*N);
        mobileparticleids.resize(nmobile);
        mpidscopy.resize(nmobile);
        immobileparticleids.resize(nmobile);
        immpidscopy.resize(nmobile);
        sizecounts.resize(nmobile);
        immpsizecounts.resize(nmobile);
        RgSqVsSize.resize(nmobile);
        immpRgSqVsSize.resize(nmobile);
        ispartofstring.resize(N);
    }
    return;
}



// Identify the most-mobile particles -- sorts from most to least mobile
// See https://stackoverflow.com/questions/25921706/creating-a-vector-of-indices-of-a-sorted-vector
struct IdxCompare {
    const std::vector<double>& target;
    IdxCompare(const std::vector<double>& target): target(target) {}
    bool operator()(int a, int b) const { return target[b] < target[a]; }
};

void idmobileparticles() {
    int i, j, k;
    pindices.clear();
    for (j = 0; j < N; ++j) pindices.push_back(j);
    sort(pindices.begin(),pindices.end(),IdxCompare(maxparticledisps)); // Altered to follow Donati and use maximum rather than current particle disps
    for (i = 0; i < nmobile; i++) {
        mpidscopy[i] = pindices[i];     mobileparticleids[i] = mpidscopy[i];
    }
    return;
}



// Same as above but sorts from lowest to highest mobility
struct IdxCompare2 {
    const std::vector<double>& target;
    IdxCompare2(const std::vector<double>& target): target(target) {}
    bool operator()(int a, int b) const { return target[b] > target[a]; }
};

void idimmobileparticles() {
    int i, j, k;
    pindices.clear();
    for (j = 0; j < N; ++j) pindices.push_back(j);
    sort(pindices.begin(),pindices.end(),IdxCompare2(maxparticledisps));  // Altered to follow Donati and use maximum rather than current particle disps
    for (i = 0; i < nmobile; i++) {
        immpidscopy[i] = pindices[i];     immobileparticleids[i] = immpidscopy[i];
    }
    return;
}




// Use Boost Graph Library to identify the mobile-particle clusters; see https://stackoverflow.com/questions/47909707/find-connected-components-using-boost-graph-library-with-the-vertex-and-edge-ty
// See https://www.boost.org/doc/libs/1_79_0/libs/graph/example/connected_components.cpp
// Defines mobile particle clusters based on mobile atoms' position at t = 0 (at step nmsdstart) rather than the current
// time t = t', following Donati et al (PRE 1999)
// Several parts of this routine still need to be parallelized
void findmobileparticleclusters() {
    int i, j, k, l, q, r, v;
    double delc[d], delc0[d], delsq, del;
    double sigmaqr, rangeqr, rangeqrsq, Rg2thiscluster;
    Graph adjmat; // mobile particles' adjacency matrix
    for (i = 0; i < nmobile; i++) add_vertex(adjmat);
// Contacts between mobile particles allow cluster identification
    vector< vector<int> > mobclist;    mobclist.resize(nmobile);

    #pragma omp parallel for schedule(guided) private(j,q,r,v,rangeqr,sigmaqr,delsq,delc0,delc)
    for (i = 0; i < nmobile-1; i++) {
        q = mpidscopy[i];
        for (v = 0; v < d; v++) delc0[v] = -Rinit[d*q+v] - Dinit*initimg[d*q+v];
        for (j = i+1; j < nmobile; j++) {
            r = mpidscopy[j];
// Apply mixing rules for particle diameters
            sigmaqr = forces.getsigmaij(q,r);
            rangeqr = adjcutoff*sigmaqr;
            delsq = 0;
            for (v = 0; v < d; v++) {
                delc[v] = delc0[v] + Rinit[d*r+v] + Dinit*initimg[d*r+v];
// Need apply min-image convention recursively here because the initial image flags of particles q, r can be different
                do { if (delc[v] < -Dinit/2.0) delc[v] += Dinit;    else if (delc[v] > Dinit/2.0) delc[v] -= Dinit; } while (delc[v] < -Dinit/2.0 || delc[v] > Dinit/2.0);
                delsq += delc[v]*delc[v];
            }
            if (delsq < rangeqr*rangeqr) mobclist[i].emplace_back(j);
        }
    }

// Build the mobile-particle contact graph
// Would be faster to implement this in parallel; see https://www.boost.org/doc/libs/1_77_0/libs/graph_parallel/doc/html/connected_components.html
// Note we have to use the COPY of mobileparticleids to avoid destroying essential info
    for (i = 0; i < nmobile-1; i++) for (j = 0; j < mobclist[i].size(); j++) add_edge(i,mobclist[i][j],adjmat);
    numclusters = connected_components(adjmat,&mpidscopy[0]);
    avgmobileclustersize = (1.0*nmobile)/numclusters;
    adjmat.clear();
    
// Recording which particles belong to each cluster allows calculating of cluster size statistics a la Figs. 5-6 of Starr/Douglas 2013
    vector <int> clustersizes;    clustersizes.resize(numclusters);
    for (i = 0; i < numclusters; i++) clustersizes[i] = 0;
    vector <int>::size_type z;
    for (z = 0; z != mpidscopy.size(); ++z) clustersizes[mpidscopy[z]]++;
    vector< vector<int> > pincluster;   pincluster.resize(numclusters);
    vector <int> clusterpos;    clusterpos.resize(numclusters);
    for (i = 0; i < nmobile; i++) {     sizecounts[i] = 0;     RgSqVsSize[i] = 0;  }
    for (i = 0; i < numclusters; i++) {
        sizecounts[clustersizes[i]]++;
        pincluster[i].resize(clustersizes[i]);
    }
    for (i = 0; i < nmobile; i++) {  // pindices and pincluster store which particles belong to which cluster
        pincluster[mpidscopy[i]][clusterpos[mpidscopy[i]]] = pindices[i];
        clusterpos[mpidscopy[i]]++;
    }
    
// Calculate the R_g-distribs -- need use the relative image flags (D.S. Simmons' method)
// and unwrapped-cluster coords to avoid spurious effects from PBCs/min-img. conv.
    vector< vector<double> > unwrappedthisclust;
    vector<double> centerpos;       centerpos.resize(d);
    vector< vector<short int> > absimgflags;
    vector< vector<bool> > clustadjmat;
    vector<bool> visited;
    int numvisited;
    double RGwr, RGuw;
    vector<bool> thisleveltouched, nextleveltouch;
    for (i = 0; i < numclusters; i++) if (clustersizes[i] > 1 && clustersizes[i] < 101) {
        unwrappedthisclust.resize(clustersizes[i]);
        absimgflags.resize(clustersizes[i]);
        clustadjmat.resize(clustersizes[i]);
        visited.resize(clustersizes[i]);
        thisleveltouched.resize(clustersizes[i]);
        nextleveltouch.resize(clustersizes[i]);
        for (j = 0; j < clustersizes[i]; j++) {
            unwrappedthisclust[j].resize(d);
            absimgflags[j].resize(d);
            clustadjmat[j].resize(clustersizes[i]);
        }
// Get the cluster adjacency matrices
        for (j = 0; j < clustersizes[i]-1; j++) for (k = j+1; k < clustersizes[i]; k++) {
            sigmaqr = forces.getsigmaij(pincluster[i][j],pincluster[i][k]);
            rangeqr = adjcutoff*sigmaqr;    // Apply mixing rules for particle diameters
            delsq = 0;
            for (v = 0; v < d; v++) {
                delc[v] = Rinit[d*pincluster[i][k] + v] + Dinit*initimg[d*pincluster[i][k] + v] - Rinit[d*pincluster[i][j] + v] - Dinit*initimg[d*pincluster[i][j] + v];
// Need apply min-image convention recursively here because the initial image flags of particles j, k can be different
                do { if (delc[v] < -Dinit/2.0) delc[v] += Dinit;    else if (delc[v] > Dinit/2.0) delc[v] -= Dinit; } while (delc[v] < -Dinit/2.0 || delc[v] > Dinit/2.0);
                delsq += delc[v]*delc[v];
            }
            if (delsq < rangeqr*rangeqr) {  clustadjmat[j][k] = true;  clustadjmat[k][j] = true; }
        }
// Recursively unwrap coordinates using breadth-first searches of the cluster adjacency-matrix graphs
// The image flags should NOT be used here because the unwrapping occurs further down
        for (j = 0; j < clustersizes[i]; j++) for (v = 0; v < d; v++) unwrappedthisclust[j][v] =  Rinit[d*pincluster[i][j] + v];
        nextleveltouch[0] = true;
        numvisited = 0;
    do {
        for (j = 0; j < clustersizes[i]; j++) if (nextleveltouch[j]) {
            for (k = 0; k < clustersizes[i]; k++) if (clustadjmat[j][k] && !visited[k]) {
                sigmaqr = forces.getsigmaij(pincluster[i][j],pincluster[i][k]);
                rangeqr = adjcutoff*sigmaqr;
                for (v = 0; v < d; v++) {
                    delc[v] = unwrappedthisclust[k][v] - unwrappedthisclust[j][v];
                    if (delc[v] < -Dinit + rangeqr) {   unwrappedthisclust[k][v] += Dinit;      absimgflags[k][v] += 1;    }
                    else if (delc[v] > Dinit - rangeqr) {  unwrappedthisclust[k][v] -= Dinit;      absimgflags[k][v] -= 1;    }
                }
                thisleveltouched[k] = true;
            }
            if (!visited[j]) {  visited[j] = true;  numvisited++;  }
            nextleveltouch[j] = false;
        }
        for (j = 0; j < clustersizes[i]; j++) {
            nextleveltouch[j] = thisleveltouched[j];
            thisleveltouched[j] = false;
        }
    }  while (numvisited < clustersizes[i]);
        
// For debugging only -- compare the wrapped and unwrapped R_g
/*        RGwr = 0;       RGuw = 0;
        vector< vector<double> > distwr, distuw;
        distwr.resize(clustersizes[i]);     distuw.resize(clustersizes[i]);
        for (j = 0; j < clustersizes[i]; j++) {     distwr[j].resize(clustersizes[i]);    distuw[j].resize(clustersizes[i]);    }
        for (j = 0; j < clustersizes[i]-1; j++) for (k = j+1; k < clustersizes[i]; k++) {
            delsq = 0;
            for (v = 0; v < d; v++) {
                delc[v] =  Rinit[d*pincluster[i][k] + v] + Dinit*initimg[d*pincluster[i][k] + v] - Rinit[d*pincluster[i][j] + v] - Dinit*initimg[d*pincluster[i][j] + v];
                do { if (delc[v] < -Dinit/2.0) delc[v] += Dinit;    else if (delc[v] > Dinit/2.0) delc[v] -= Dinit; } while (delc[v] < -Dinit/2.0 || delc[v] > Dinit/2.0);
                delsq += delc[v]*delc[v];
            }
            RGwr += delsq/(clustersizes[i]*clustersizes[i]);
            distwr[j][k] = sqrt(delsq);     distwr[k][j] = sqrt(delsq);
            delsq = 0;
            for (v = 0; v < d; v++) {
                delc[v] = unwrappedthisclust[k][v] - unwrappedthisclust[j][v];
                delsq += delc[v]*delc[v];
            }
            RGuw += delsq/(clustersizes[i]*clustersizes[i]);
            distuw[j][k] = sqrt(delsq);     distuw[k][j] = sqrt(delsq);
        }
        if (fabs(sqrt(RGuw/RGwr) - 1) > .01 && clustersizes[i] < 20) {
            cout << "Cluster " << i << " size = " << clustersizes[i] << endl;
            cout << "Wrapped coords:" << endl;
            for (j = 0; j < clustersizes[i]; j++) {
                for (v = 0; v < d; v++) cout << setw(11) << Rinit[d*pincluster[i][j] + v];
                cout << endl;
            }
            cout << "Wrapped R_g = " << sqrt(RGwr) << endl;
            cout << "Wrapped distance matrix:" << endl;
            for (j = 0; j < clustersizes[i]; j++) {
                for (k = 0; k < clustersizes[i]; k++) cout << setw(10) << distwr[j][k];
                cout << endl;
            }
            cout << "Unwrapped coords:" << endl;
            for (j = 0; j < clustersizes[i]; j++) {
                for (v = 0; v < d; v++) cout << setw(11) << unwrappedthisclust[j][v];
                cout << endl;
            }
            cout << "Cluster adj-mat:" << endl;
            for (j = 0; j < clustersizes[i]; j++) {
                for (k = 0; k < clustersizes[i]; k++) cout << setw(3) << clustadjmat[j][k];
                cout << endl;
            }
            cout << "Intracluster image flags:" << endl;
            for (j = 0; j < clustersizes[i]; j++) {
                for (k = 0; k < d; k++) cout << setw(3) << absimgflags[j][k];
                cout << endl;
            }
            cout << "visited: ";
            for (j = 0; j < clustersizes[i]; j++) cout << visited[j] << " ";
            cout << endl;
            cout << "nextleveltouch: ";
            for (j = 0; j < clustersizes[i]; j++) cout << nextleveltouch[j] << " ";
            cout << endl;
            cout << "Unwrapped R_g = " << sqrt(RGuw) << " : unwrapped/wrapped ratio = " << sqrt(RGuw/RGwr) << endl;
            cout << "Unwrapped distance matrix:" << endl;
            for (j = 0; j < clustersizes[i]; j++) {
                for (k = 0; k < clustersizes[i]; k++) cout << setw(10) << distuw[j][k];
                cout << endl;
            }
        }
        distwr.clear();     distuw.clear();  */

// Get this cluster's mean position using the UNWRAPPED coordinates, then calculate its R_g^2
        for (v = 0; v < d; v++) centerpos[v] = 0;
        for (j = 0; j < clustersizes[i]; j++) for (v = 0; v < d; v++) centerpos[v] += unwrappedthisclust[j][v]/clustersizes[i];
        Rg2thiscluster = 0;
        for (j = 0; j < clustersizes[i]; j++) for (v = 0; v < d; v++) {
            delc[v] = unwrappedthisclust[j][v] - centerpos[v];
            Rg2thiscluster += delc[v]*delc[v]/clustersizes[i];
        }
        RgSqVsSize[clustersizes[i]] += Rg2thiscluster/sizecounts[clustersizes[i]];
        unwrappedthisclust.clear();
        absimgflags.clear();
        clustadjmat.clear();
        visited.clear();
        thisleveltouched.clear();
        nextleveltouch.clear();
    } // end of the loop over clusters
// output the results of the size- and Rg-distrib calculations
    readwrite.writemobilestats();
    
    return;
}



// Same as above, but for the LEAST-mobile particles
// Several parts of this routine still need to be parallelized
void findimmobileparticleclusters() {
    int i, j, k, l, q, r, v;
    double delc[d], delc0[d], delsq, del;
    double sigmaqr,rangeqr, rangeqrsq, Rg2thiscluster;
    Graph adjmat; // mobile particles' adjacency matrix
    for (i = 0; i < nmobile; i++) add_vertex(adjmat);
// Contacts between mobile particles allow cluster identification
    vector< vector<int> > immobclist;    immobclist.resize(nmobile);

    #pragma omp parallel for schedule(guided) private(j,q,r,v,rangeqr,sigmaqr,delsq,delc0,delc)
    for (i = 0; i < nmobile-1; i++) {
        q = immpidscopy[i];
        for (v = 0; v < d; v++) delc0[v] = -Rinit[d*q+v] - Dinit*initimg[d*q+v];
        for (j = i+1; j < nmobile; j++) {
            r = immpidscopy[j];
// Apply mixing rules for particle diameters
            sigmaqr = forces.getsigmaij(q,r);
            rangeqr = adjcutoff*sigmaqr;
            delsq = 0;
            for (v = 0; v < d; v++) {
                delc[v] = delc0[v] + Rinit[d*r+v] + Dinit*initimg[d*r+v];
// Need apply min-image convention recursively here because the initial image flags of particles q, r can be different
                do { if (delc[v] < -Dinit/2.0) delc[v] += Dinit;    else if (delc[v] > Dinit/2.0) delc[v] -= Dinit; } while (delc[v] < -Dinit/2.0 || delc[v] > Dinit/2.0);
                delsq += delc[v]*delc[v];
            }
            if (delsq < rangeqr*rangeqr) immobclist[i].emplace_back(j);
        }
    }

// Build the immobile-particle contact graph
// Would be faster to implement this in parallel; see https://www.boost.org/doc/libs/1_77_0/libs/graph_parallel/doc/html/connected_components.html
    for (i = 0; i < nmobile-1; i++) for (j = 0; j < immobclist[i].size(); j++) add_edge(i,immobclist[i][j],adjmat);
    numimmclusters = connected_components(adjmat,&immpidscopy[0]);
    avgimmobileclustersize = (1.0*nmobile)/numimmclusters;
    adjmat.clear();
    
// Recording which particles belong to each cluster allows calculating of cluster size statistics a la Figs. 5-6 of Starr/Douglas 2013
    vector <int> clustersizes;    clustersizes.resize(numimmclusters);
    for (i = 0; i < numimmclusters; i++) clustersizes[i] = 0;
    vector <int>::size_type z;
    for (z = 0; z != immpidscopy.size(); ++z) clustersizes[immpidscopy[z]]++;
    vector< vector<int> > pincluster;   pincluster.resize(numimmclusters);
    vector <int> clusterpos;    clusterpos.resize(numimmclusters);
    for (i = 0; i < nmobile; i++) {     immpsizecounts[i] = 0;     immpRgSqVsSize[i] = 0;  }
    for (i = 0; i < numimmclusters; i++) {
        immpsizecounts[clustersizes[i]]++;
        pincluster[i].resize(clustersizes[i]);
    }
    for (i = 0; i < nmobile; i++) {  // pindices and pincluster store which particles belong to which cluster
        pincluster[immpidscopy[i]][clusterpos[immpidscopy[i]]] = pindices[i];
        clusterpos[immpidscopy[i]]++;
    }

// Calculate the R_g-distribs -- need use the relative image flags (D.S. Simmons' method)
// and unwrapped-cluster coords to avoid spurious effects from PBCs/min-img. conv.
    vector< vector<double> > unwrappedthisclust;
    vector<double> centerpos;       centerpos.resize(d);
    vector< vector<short int> > absimgflags;
    vector< vector<bool> > clustadjmat;
    vector<bool> visited;
    int numvisited;
    vector<bool> thisleveltouched, nextleveltouch;
    for (i = 0; i < numimmclusters; i++) if (clustersizes[i] > 1 && clustersizes[i] < 101) {
        unwrappedthisclust.resize(clustersizes[i]);
        absimgflags.resize(clustersizes[i]);
        clustadjmat.resize(clustersizes[i]);
        visited.resize(clustersizes[i]);
        thisleveltouched.resize(clustersizes[i]);
        nextleveltouch.resize(clustersizes[i]);
        for (j = 0; j < clustersizes[i]; j++) {
            unwrappedthisclust[j].resize(d);
            absimgflags[j].resize(d);
            clustadjmat[j].resize(clustersizes[i]);
        }
    // Get the cluster adjacency matrices
        for (j = 0; j < clustersizes[i]-1; j++) for (k = j+1; k < clustersizes[i]; k++) {
            sigmaqr = forces.getsigmaij(pincluster[i][j],pincluster[i][k]);
            rangeqr = adjcutoff*sigmaqr;    // Apply mixing rules for particle diameters
            delsq = 0;
            for (v = 0; v < d; v++) {
                delc[v] = Rinit[d*pincluster[i][k] + v] + Dinit*initimg[d*pincluster[i][k] + v] - Rinit[d*pincluster[i][j] + v] - Dinit*initimg[d*pincluster[i][j] + v];
// Need apply min-image convention recursively here because the initial image flags of particles j, k can be different
                do { if (delc[v] < -Dinit/2.0) delc[v] += Dinit;    else if (delc[v] > Dinit/2.0) delc[v] -= Dinit; } while (delc[v] < -Dinit/2.0 || delc[v] > Dinit/2.0);
                delsq += delc[v]*delc[v];
            }
            if (delsq < rangeqr*rangeqr) {  clustadjmat[j][k] = true;  clustadjmat[k][j] = true; }
        }
// Recursively unwrap coordinates using breadth-first searches of the cluster adjacency-matrix graphs
// The image flags should NOT be used here because the unwrapping occurs further down
        for (j = 0; j < clustersizes[i]; j++) for (v = 0; v < d; v++) unwrappedthisclust[j][v] =  Rinit[d*pincluster[i][j] + v];
        nextleveltouch[0] = true;
        numvisited = 0;
    do {
        for (j = 0; j < clustersizes[i]; j++) if (nextleveltouch[j]) {
            for (k = 0; k < clustersizes[i]; k++) if (clustadjmat[j][k] && !visited[k]) {
                sigmaqr = forces.getsigmaij(pincluster[i][j],pincluster[i][k]);
                rangeqr = adjcutoff*sigmaqr;
                for (v = 0; v < d; v++) {
                    delc[v] = unwrappedthisclust[k][v] - unwrappedthisclust[j][v];
                    if (delc[v] < -Dinit + rangeqr) {   unwrappedthisclust[k][v] += Dinit;      absimgflags[k][v] += 1;    }
                    else if (delc[v] > Dinit - rangeqr) {  unwrappedthisclust[k][v] -= Dinit;      absimgflags[k][v] -= 1;    }
                }
                thisleveltouched[k] = true;
            }
            if (!visited[j]) {  visited[j] = true;  numvisited++;  }
            nextleveltouch[j] = false;
        }
        for (j = 0; j < clustersizes[i]; j++) {
            nextleveltouch[j] = thisleveltouched[j];
            thisleveltouched[j] = false;
        }
    }  while (numvisited < clustersizes[i]);
// Get this cluster's mean position using the UNWRAPPED coordinates, then calculate its R_g^2
        for (v = 0; v < d; v++) centerpos[v] = 0;
        for (j = 0; j < clustersizes[i]; j++) for (v = 0; v < d; v++) centerpos[v] += unwrappedthisclust[j][v]/clustersizes[i];
        Rg2thiscluster = 0;
        for (j = 0; j < clustersizes[i]; j++) for (v = 0; v < d; v++) {
            delc[v] = unwrappedthisclust[j][v] - centerpos[v];
            Rg2thiscluster += delc[v]*delc[v]/clustersizes[i];
        }
        immpRgSqVsSize[clustersizes[i]] += Rg2thiscluster/immpsizecounts[clustersizes[i]];
        unwrappedthisclust.clear();
        absimgflags.clear();
        clustadjmat.clear();
        visited.clear();
        thisleveltouched.clear();
        nextleveltouch.clear();
    } // end of the loop over clusters

// Output the results of the size- and Rg-distrib calculations
    readwrite.writeimmobilestats();
    
    return;
}



// Find the strings using the criteria that mobile particle i is part of a string if it is within half
// the diameter of mobile particle j (of mobile particle j's initial position)
// Several parts of this routine still need to be parallelized
void findstrings() {
    int i, j, k, q, r, v;
    double delc0[d], delc[d], delsq;
    double sigmaqr, rangeqr, rangeqrsq, Rg2thisstring;
// This cutoff criterion (rfrac) gives r = half the particle diameter (the usual value) in 3D and preserves the
// replacement-within-the-inner-1/8-of-the-original-occupied-volume criterion in higher d
    double rfrac = pow(1.0/8,1.0/d);
    double dsqcutcoeff;
    if (model == 1 || model == 2) dsqcutcoeff = rfrac*rfrac;
    else if (model == 3) dsqcutcoeff = rfrac*rfrac*TWO16*TWO16;

// First identify the particles that form the strings
    for (i = 0; i < N; i++) ispartofstring[i] = false;
    ninstrings = 0;
    #pragma omp parallel for schedule(guided) private(j,v,delc,delsq) reduction(+:ninstrings)
    for (i = 0; i < nmobile; i++) {
        for (j = 0; j < nmobile; j++) {
            if (j == i) continue;
            delsq = 0;
            for (v = 0; v < d; v++) {
                delc[v] = RVF[3*d*mobileparticleids[i] + v] + image[d*mobileparticleids[i] + v]*D - Rinit[d*mobileparticleids[j] + v] - initimg[d*mobileparticleids[j] + v]*Dinit;
// Need apply min-image convention recursively here because the initial image flags of particles i, j can be different
                do { if (delc[v] < -D/2.0) delc[v] += D;    else if (delc[v] > D/2.0) delc[v] -= D; } while (delc[v] < -D/2.0 || delc[v] > D/2.0);
                delsq += delc[v]*delc[v];
            }
// Below criterion is consistent with Starr/Douglas/Sastry 2013: position-replacement distance cutoff = 2^(-5/6) for standard LJ
// Use convention that each particle can be part of at most one string, but add both particles i AND j to strings, a la Zhang/Douglas 2015
            if (delsq < dsqcutcoeff*sigma[mobileparticleids[j]]*sigma[mobileparticleids[j]]) {
                if (!ispartofstring[mobileparticleids[i]]) {    ispartofstring[mobileparticleids[i]] = true;    ninstrings++;  }
                if (!ispartofstring[mobileparticleids[j]]) {    ispartofstring[mobileparticleids[j]] = true;    ninstrings++;  }
            }
        }
    }
    stringfrac = (1.0*ninstrings)/nmobile;
    
// Next gather the length- and R_g-distributions for strings of length up to 100
    stringlengthcounts.resize(max(ninstrings,101));
    StringRgSqVsLength.resize(max(ninstrings,101));
    for (i = 0; i < 101; i++) { stringlengthcounts[i] = 0;  StringRgSqVsLength[i] = 0; }
    avgstringsize = 0;
    
// Here duplicate the procedure used to ID the mobile-particle clusters, but for strings
if (ninstrings > 0) {
    vector <int> stringidscopy, stringparticleids;
    stringidscopy.resize(ninstrings);
    stringparticleids.resize(ninstrings);
// Store the string-particle IDs so can build the strings' adjacency-matrix graphs
    j = 0;
    for (i = 0; i < N; i++) if (ispartofstring[i]) {
        stringidscopy[j] = i;
        stringparticleids[j] = stringidscopy[j];
        j++;
    }
// Build the strings' adjacency-matrix graphs
    vector< vector<int> > stringclist;    stringclist.resize(ninstrings);
    Graph adjmat; // mobile particles' adjacency matrix
    for (i = 0; i < ninstrings; i++) add_vertex(adjmat);
    #pragma omp parallel for schedule(guided) private(j,q,r,v,rangeqr,sigmaqr,delsq,delc0,delc)
    for (i = 0; i < ninstrings-1; i++) {
        q = stringidscopy[i];
        for (v = 0; v < d; v++) delc0[v] = -RVF[3*d*q+v];
        for (j = i+1; j < ninstrings; j++) {
            r = stringidscopy[j];
// Apply mixing rules for particle diameters
            sigmaqr = forces.getsigmaij(q,r);
            rangeqr = adjcutoff*sigmaqr;
            delsq = 0;
            for (v = 0; v < d; v++) {
                delc[v] = delc0[v] + RVF[3*d*r+v];
                if (delc[v] < -D/2.0) delc[v] += D;    else if (delc[v] > D/2.0) delc[v] -= D;
                delsq += delc[v]*delc[v];
            }
            if (delsq < rangeqr*rangeqr) stringclist[i].emplace_back(j);
        }
    }
    for (i = 0; i < ninstrings-1; i++) for (j = 0; j < stringclist[i].size(); j++) add_edge(i,stringclist[i][j],adjmat);
    numstrings = connected_components(adjmat,&stringidscopy[0]);
    adjmat.clear();
    avgstringsize = (1.0*ninstrings)/numstrings;

    
// Recording which particles belong to each string allows calculating string length statistics a la Figs. 5-6 of Starr/Douglas 2013
    vector <int> stringlengths;    stringlengths.resize(numstrings);
    vector <int>::size_type z;
    for (z = 0; z != stringidscopy.size(); ++z) stringlengths[stringidscopy[z]]++;
    vector< vector<int> > pinstring;   pinstring.resize(numstrings);
    vector <int> stringpos;    stringpos.resize(numstrings);
    for (i = 0; i < numstrings; i++) {
        stringlengthcounts[stringlengths[i]]++;
        pinstring[i].resize(stringlengths[i]);
    }
    for (i = 0; i < ninstrings; i++) {  // analogue of pindices and pincluster in above routines
        pinstring[stringidscopy[i]][stringpos[stringidscopy[i]]] = stringparticleids[i];
        stringpos[stringidscopy[i]]++;
    }
    
// Calculate the R_g-distribs using a procedure analogous to that used above for mobile particles
    vector< vector<double> > unwrappedthisstring;
    vector<double> centerpos;       centerpos.resize(d);
    vector< vector<short int> > absimgflags;
    vector< vector<bool> > stringadjmat;
    vector<bool> visited;
    int numvisited;
    vector<bool> thisleveltouched, nextleveltouch;
    for (i = 0; i < numstrings; i++) if (stringlengths[i] > 1 && stringlengths[i] < 101) {
        unwrappedthisstring.resize(stringlengths[i]);
        absimgflags.resize(stringlengths[i]);
        stringadjmat.resize(stringlengths[i]);
        visited.resize(stringlengths[i]);
        thisleveltouched.resize(stringlengths[i]);
        nextleveltouch.resize(stringlengths[i]);
        for (j = 0; j < stringlengths[i]; j++) {
            unwrappedthisstring[j].resize(d);
            absimgflags[j].resize(d);
            stringadjmat[j].resize(stringlengths[i]);
        }
        for (j = 0; j < stringlengths[i]-1; j++) for (k = j+1; k < stringlengths[i]; k++) {
            sigmaqr = forces.getsigmaij(pinstring[i][j],pinstring[i][k]);
            rangeqr = adjcutoff*sigmaqr;
            delsq = 0;
            for (v = 0; v < d; v++) {
                delc[v] = RVF[3*d*pinstring[i][k] + v] - RVF[3*d*pinstring[i][j] + v];
                if (delc[v] < -D/2.0) delc[v] += D;    else if (delc[v] > D/2.0) delc[v] -= D;
                delsq += delc[v]*delc[v];
            }
            if (delsq < rangeqr*rangeqr) {  stringadjmat[j][k] = true;  stringadjmat[k][j] = true; }
        }
        for (j = 0; j < stringlengths[i]; j++) for (v = 0; v < d; v++) unwrappedthisstring[j][v] = RVF[3*d*pinstring[i][j] + v];
        nextleveltouch[0] = true;
        numvisited = 0;
// Recursively unwrap coordinates using a proper breadth-first search a la D. Simmons' suggestion
    do {
        for (j = 0; j < stringlengths[i]; j++) if (nextleveltouch[j]) {
            for (k = 0; k < stringlengths[i]; k++) if (stringadjmat[j][k] && !visited[k]) {
                sigmaqr = forces.getsigmaij(pinstring[i][j],pinstring[i][k]);
                rangeqr = adjcutoff*sigmaqr;
                for (v = 0; v < d; v++) {
                    delc[v] = unwrappedthisstring[k][v] - unwrappedthisstring[j][v];
                    if (delc[v] < -D + rangeqr) {   unwrappedthisstring[k][v] += D;      absimgflags[k][v] += 1;    }
                    else if (delc[v] > D - rangeqr) {  unwrappedthisstring[k][v] -= D;      absimgflags[k][v] -= 1;    }
                }
                thisleveltouched[k] = true;
            }
            if (!visited[j]) {  visited[j] = true;  numvisited++;  }
            nextleveltouch[j] = false;
        }
        for (j = 0; j < stringlengths[i]; j++) {
            nextleveltouch[j] = thisleveltouched[j];
            thisleveltouched[j] = false;
        }
      }  while (numvisited < stringlengths[i]);
// Get this string's mean position using the UNWRAPPED coordinates, then calculate its R_g^2
        for (v = 0; v < d; v++) centerpos[v] = 0;
        for (j = 0; j < stringlengths[i]; j++) for (v = 0; v < d; v++) centerpos[v] += unwrappedthisstring[j][v]/stringlengths[i];
        Rg2thisstring = 0;
        for (j = 0; j < stringlengths[i]; j++) for (v = 0; v < d; v++) {
            delc[v] = unwrappedthisstring[j][v] - centerpos[v];
            Rg2thisstring += delc[v]*delc[v]/stringlengths[i];
        }
        if (stringlengths[i] < 101) {
            StringRgSqVsLength[stringlengths[i]] += Rg2thisstring/stringlengthcounts[stringlengths[i]];
        }
        unwrappedthisstring.clear();
        absimgflags.clear();
        stringadjmat.clear();
        visited.clear();
        thisleveltouched.clear();
        nextleveltouch.clear();
    } // end of R_g-calculation loop
  } // end of if (ninstrings > 0) conditional

// Output the results of the length- and Rg-distrib calculations
    readwrite.writestringstats();
    return;
}
