 // forcelaw.cpp -- force and energy evaluation
#include"forcelaw.h"
// Global variables

// sigmaij (particle-diameter mixing rule)
double forcelaw::getsigmaij(int i, int u) {
    double sigmaij;
    switch (model) {
        case 1: // generalized-LJ
            sigmaij = (sigma[i]+sigma[u])/2;
            break;
        case 2: // Morse
            sigmaij = (sigma[i]+sigma[u])/2;
            break;
        case 3: // Kob-Andersen
            if (i < Nlarge && u < Nlarge) sigmaij = 1.0*sAA; // A-A pairs
            else if (i >= Nlarge && u >= Nlarge) sigmaij = 0.88*sAA; // B-B pairs
            else sigmaij = 0.8*sAA; // A-B pairs
            break;
    }
    return sigmaij;
}

// epsilonij (particle-energy-prefactor mixing rule)
double forcelaw::getepsilonij(int i, int u) {
    double epsilonij;
    switch (model) {
        case 1: // generalized-LJ
            epsilonij = sqrtepsilon[i]*sqrtepsilon[u];
            break;
        case 2: // Morse
            epsilonij = sqrtepsilon[i]*sqrtepsilon[u];
            break;
        case 3: // Kob-Andersen
            if (i < Nlarge && u < Nlarge) epsilonij = 1.0;  // A-A pairs
            else if (i >= Nlarge && u >= Nlarge) epsilonij = 0.5; // B-B pairs
            else epsilonij = 1.5; // A-B pairs
            break;
    }
    return epsilonij;
}

// Alt. version for SWAP
// sigmaij (particle-diameter mixing rule)
double forcelaw::getsigmaij2(int i, int u) {
    double sigmaij;
    switch (model) {
        case 1: // generalized-LJ
            sigmaij = (sigma[i]+sigma[u])/2;
            break;
        case 2: // Morse
            sigmaij = (sigma[i]+sigma[u])/2;
            break;
        case 3: // Kob-Andersen
            if (sigma[i] == sAA && sigma[u] == sAA) sigmaij = 1.0*sAA; // A-A pairs
            else if (sigma[i] == 0.88*sAA && sigma[u] == 0.88*sAA) sigmaij = 0.88*sAA; // B-B pairs
            else sigmaij = 0.8*sAA; // A-B pairs
            break;
    }
    return sigmaij;
}

// Alt. version for SWAP
// epsilonij (particle-energy-prefactor mixing rule)
double forcelaw::getepsilonij2(int i, int u) {
    double epsilonij;
    switch (model) {
        case 1: // generalized-LJ
            epsilonij = sqrtepsilon[i]*sqrtepsilon[u];
            break;
        case 2: // Morse
            epsilonij = sqrtepsilon[i]*sqrtepsilon[u];
            break;
        case 3: // Kob-Andersen
            if (sigma[i] == sAA && sigma[u] == sAA) epsilonij = 1.0;  // A-A pairs
            else if (sigma[i] == 0.88*sAA && sigma[u] == 0.88*sAA) epsilonij = 0.5; // B-B pairs
            else epsilonij = 1.5; // A-B pairs
            break;
    }
    return epsilonij;
}


// F/r (force/distance)
double forcelaw::getfovr(double delsq, double sigmaij, double sigmaijsq, double epsilonij) {
    double del, facij, fovr;
    switch (model) {
        case 1: // generalized-LJ
            facij = mypow(sigmaijsq/delsq);
            fovr = 2*n*epsilonij*facij*(facij-1)/delsq;
            break;
        case 2: // Morse
//            EXPA = exp(alpha);
            del = sqrt(delsq/sigmaijsq);
            facij = EXPA*exp(-alpha*del);
            fovr = (2*alpha*epsilonij/(del*sigmaij))*facij*(facij-1);
            break;
        case 3: // Kob-Andersen
            facij = mypow(sigmaijsq/delsq);
            fovr = 4*n*epsilonij*facij*(2*facij-1)/delsq;
            break;
    }
    return fovr;
}

// pair energy
double forcelaw::getepair(double delsq, double sigmaij, double sigmaijsq, double epsilonij) {
//    double del, EXPA, facij, facijtmp, shift, ep;
    double del, facij, shift, ep;
    switch (model) {
        case 1: // generalized-LJ
//            facijtmp = mypow(1.0/cutsq);
            shift = epsilonij*facijtmp*(facijtmp-2);
            facij = mypow(sigmaijsq/delsq);
            ep = epsilonij*facij*(facij-2) - shift;
            break;
        case 2: // Morse
//            EXPA = exp(alpha);
//            facijtmp = EXPA*exp(-alpha*cut);
            shift = epsilonij*facijtmp*(facijtmp-2);
            del = sqrt(delsq/sigmaijsq);
            facij = EXPA*exp(-alpha*del);
            ep = epsilonij*facij*(facij-2) - shift;
            break;
        case 3: // Kob-Andersen
//            facijtmp = mypow(1.0/cutsq);
            shift = 4*epsilonij*facijtmp*(facijtmp-1);
            facij = mypow(sigmaijsq/delsq);
            ep = 4*epsilonij*facij*(facij-1) - shift;
            break;
    }
    return ep;
}


// For force & energy calcs. for generalized-LJ:
// This version works and is optimized for arbitrary EVEN n
double forcelaw::mypow(double x) {
    double y = 1.0;
    int j = n/2;  // this edit of mypow2 assumes n is even!!
    int i;
    for (;;) {
        if (j&1) y *= x;
        j >>= 1;
        if (!j) break;
        x *= x;
    }
    return y;
}


// Set the MD timestep
double forcelaw::setdt() {
    double thisdt;
    switch (model) {
        case 1: // generalized-LJ
            thisdt = 6.0/(125*n);
            break;
        case 2: // Morse
            thisdt = 6.0/(125*alpha);
            break;
        case 3: // Kob-Andersen
            thisdt = 6.0/(125*n);
            break;
    }
    return thisdt;
}



// Set constant factors like EXPA, facijtmp, etc.
void forcelaw::initprefacs() {
    EXPA = exp(alpha);
    switch (model) {
        case 1: // generalized-LJ
            facijtmp = mypow(1.0/cutsq);
            break;
        case 2: // Morse
            EXPA = exp(alpha);
            facijtmp = EXPA*exp(-alpha*cut);
            break;
        case 3: // Kob-Andersen
            facijtmp = mypow(1.0/cutsq);
            break;
    }
    return;
}
