#include"cellmanagement.h"
#include"SWAP.h"
#include"IO.h"
#include"diagnostics.h"
#include"utility.h"
#include"minimize.h"
#include"forcelaw.h"
#include"sched.h"
#include"string.h"
#define PI 3.1415926536

// Global variables
double* RVF;    // current positions, velocities, forces
double* Rlast;  // positions at last NL build
double* Rinit;  // initial positions (for calculation of MSD, etc.)
double* mass;   // particle masses
double* sigma;  // particle diameters
double* epsilon;    // particle energy/force prefactors
double* sqrtepsilon; // saves having to calculate sqrt(epsilon_i epsilon_j) every timestep
int* Z_array;   // array for counting Z
int* allneigh;  // neighbor list
short int* numneigh;    // number of neighbors/atom
int maxneigh;   // maximum number of neighbors/atom
short int* image;   // current particle image flags
short int* initimg; // initial image flags (for calculation of MSD, etc.)
int *nnforg, *nnforgLL, *nnforgLS, *nnforgSS;    // for calculating g(r)
double ALU[3];  // polydisperse-sigma-distribution parameters
int d;      // spatial dimension
int N;      // number of particles simulated
int Nlarge, Nsmall;  // number of A, B particles in KA bidisperse mixture
int ncell;  // number of cubic subcells along each dimension
int ncellprev;  // previous value of ncell (needed if cell volume V varies)
int totcells;   // total number of subcells
int nbs;    // nbs = 3 or 5 controls structure of cell linking
double skin;    // skin depth for NL builds
double cut, cutsq;  // for longer-range interactions
int neighlistbuilds;    // number of NL builds performed
double avgneigh, totavgneigh;   // current and overall average neighbors/particle
int nthreads;   // parallel simulations run on nthreads cores
const int nchunk = 100;     // Use schedule(static,nchunk) for integration of all EOM
double dt, dthalf;      // timestep, half-timestep for leapfrog, etc.
double elapstime;  // "Lennard-Jones" time since startstep or nmsdstart
int nsteps;     // simulation runs for nsteps timesteps
int nbet;       // leapfrog this many steps between NL-build-checks
int nthermo;    // calculate thermodynamic data every nthermo timesteps
int nwindow;    // average thermo data over time windows of length nwindow timesteps
int nvhc;       // calculate van Hove corr. function every nvhc timesteps
int ngofr;      // Calculate g(r) every ngofr timesteps
int nbaro;      // apply barostat every nbaro timesteps
int nmin;       // energy-minimize every nmin timesteps
int nswap;       // SWAP every nswap timesteps
int nmsdstart;  // start MSD calculation after nmsdstart steps
int thiswindowstart;  // timestep at start of the current "window"
int nrestart;   // write restart file every nrestart timesteps
int ndump;     // write coordinates every ndump timesteps
int nmobdump;   // write mobile-particle coords every nmobdump timesteps
int step, startstep;   // current and starting timesteps
int lastswapstep;  // sometimes required for stability when SWAP employed
int rattlers, floaters; //calculating rattler and floating particles
long long int swapatt, swapsucc;    // number of attempted & successful swaps
double attfrac;  // attempt swaps for attfrac of particles every time swapmove() is called
double D;   // simulation cell side length
double Dinit;   // initial value of D (for calculation of MSD, etc.)
double lo, hi;  // positions of cell's lower, upper boundaries (typically -/+ D/2)
double rho, phi;    // particle number density and packing fraction
double totpartvol, vol;     // total particle volume & cell volume
double E, EP, EK;  // total, pair and kinetic energies
double EXPA, facijtmp;  // for speeding up force & energy evaluation
double T, P, Z, Znoratt;    // temperature, pressure, average coordination number
double Ttarg, Ptarg;    // target temperature and pressure for thermostat, barostat
double Tdot = .00001; // for runs with temperature ramp
double msd, namsd, ngp, maxdisp, f_ov, SISF;  // measures of diffusion
double msd1, m4d1, ngp1, fov1, msd2, m4d2, ngp2, fov2;  // for bidisperse systems
double msdm, m4dm, ngpm, fovm, msdimm, m4dimm, ngpimm, fovimm;  // for mobile and immobile particles
double sisf1, sisf2, sisfm, sisfimm;  // self-intermediate scattering functions of large, small, mobile, immobile particles
double avgmobileclustersize, avgimmobileclustersize, stringfrac, avgstringsize;  // for mobile/immobile particle-cluster and mobile-particle-string analyses
double min_actual_time; // total time elapsed in FIRE and total displacement for SD

// Functions defined in main.cpp
void leapfrog();    // Leapfrog integration with thermostat
void getforce(int i);   // Get force on atom i
void Berendsen(double Ptarg);   // Berendsen thermostat

// Functions defined in other routines
extern double initRVFMSE(bool readresflag, const char *initresfname, const char *initRinitfname); // Initialize particle positions, diameters, velocities, image flags, etc.
extern void dealloc();      // deallocate all large arrays

// Class declarations
cell supercell;
swapMC SWAP;
IO readwrite;
diagnostics diags;
minimizer minims;
forcelaw forces;
utility utils;

// Parameters specific to this simulation
const int model = 2;  // 1 for generalized LJ (Mie), 2 for Morse, 3 for Kob-Andersen
int n = 6; // generalized-LJ exponent
int alpha = 30;  // Morse exponent
const double sAA = 1; // sigma for the A-particles in a Kob-Andersen simulation
bool baroflag = false;    // True if applying barostat, otherwise false
bool minflag = false;      // True if performing periodic E-minimizations
bool vhcflag = false;     // True if outputting van Hove correlation function
bool gofrflag = false;     // True if outputting pair correlation function
bool sofqflag = false;     // True if calculating self-intermediate scattering function
bool mobileflag = false;    // True if identifying mobile-particle clusters
const double tau_press = 10; // Berendsen barostat pressure damping time
const double tau_temp = 3.16228; // Berendsen thermostat temperature damping time
double dispcap = .025;      // Maximum single-step displacement cap for integration of various EOM
double paircorrcap = 10; // Maximum distance out to which to calculate g(r) -- needs to be <= D/2.
double VHCcap = 10; // Maximum distance out to which to calculate self van Hove corr. function
int grcap = floor(paircorrcap/.01);  // number of bins in the output g(r)
int vhcap = floor(VHCcap/.01);  // number of bins in the output G_s(r)


int main (int argc, const char *argv[]) {
    d = atoi(argv[1]);  // spatial dimension
    N = atoi(argv[2]);  // number of particles
    phi = atof(argv[3]);    // volume fraction
    Ttarg = atof(argv[4]);  // target temperature
    nsteps = atoi(argv[5]);     // # of steps in run
    nthreads = atoi(argv[6]);   // # of parallel threads
    attfrac = atof(argv[7]);   // fraction of atoms to attempt SWAP moves for / tau
// Can start simulation from a restart file and a file with Rinit values
// (latter so can continue calculation of msd, ngp, f_ov, etc.)
    const char *initresfname = argv[8];  // (optional) start simulation from this restart file
    const char *initRinitfname = argv[9]; // (optional) reference state for MSD calculation
    nmsdstart = atoi(argv[10]);  // start msd calculation after this many steps
    bool readresflag;
    if (!strcmp(initresfname,"null") && !strcmp(initRinitfname,"null")) readresflag = false;
    else readresflag = true;
    
// Initialize loop counters and various run-progress-related variables
    int i, j, k, l;
    long long int lastswapsucc;  // for checking whether need rebuild NL after swapping
    if (!readresflag) step = 0;   // MD timestep
    swapatt = 0;    swapsucc = 0;   // SWAP attempts and successes
// Set the MD timestep
    dt = forces.setdt();
    dthalf = dt/2;
    
// Set the number of large and small particles (e.g. the  A, B particles in a Kob-Andersen mixture)
// For continuously-polydisp it makes sense to set afrac = 1/2
    double afrac = 1.0/2;  // Std. value is 4/5 but 2/3 is a better KA glass-former
    Nlarge = round(afrac*N);
    Nsmall = N - Nlarge;
    cout << "Nlarge = " << Nlarge << " Nsmall = " << Nsmall << endl;
    
// Counters for NL-build tracking
    neighlistbuilds = 0;    totavgneigh = 0;
// Thermo/barostat and other simulation-control parameters: below outputs thermo data once/tau,
// calls barostat 25 times/tau, writes restart files once/(1000tau), etc.
    nthermo = static_cast<int>(1/dt);   // # of steps to run between calls to getthermo()
    nwindow = static_cast<int>(1000/dt);  // average over windows of length 1000τ
    nvhc = static_cast<int>(100/dt);     // # of steps to run between calls to getvanHove()
    ngofr = static_cast<int>(10000/dt);     // # of steps to run between calls to getgofr()
    nbaro = static_cast<int>(.04/dt);   // # of steps to run between calls to Berendsen()
    nrestart = static_cast<int>(10000/dt);   // # of steps between restart-file writes
    ndump = static_cast<int>(10000/dt);   // # of steps between coordinate dumps
    nmobdump = static_cast<int>(10000/dt);   // # of steps between mobile-particle coordinate dumps
    nbet = static_cast<int>(.008/dt);   // # of steps to run between NL-build checks
    nmin = static_cast<int>(10/dt);   // # of steps to run between energy minimizations
    nswap = static_cast<int>(1/dt);    // # of steps to run between SWAP MC passes
    
// Timers for high resolution timing
    chrono::duration<double> elapsed, totelapsed, nltime, totnltime, verltime, totverltime;
    chrono::duration<double> pushtime, mintime, totmintime, barotime, totbarotime;
    chrono::duration<double> swaptime, totswaptime, diagtime, totdiagtime;
    auto start = chrono::high_resolution_clock::now();
    
// Verlet-list cutoff is at r_c = cut*(max. particle diameter) + skin
    cut = 1.0;    cutsq = cut*cut;
    skin = .25;
// Set the cut-dependent constants used in force/energy evals
    forces.initprefacs();

// Initalize all per-particle quantities (positions, diameters, velocities, image flags, neighbor arrays, etc.)
    totpartvol = utils.initRVFMSE(readresflag,initresfname,initRinitfname);
// Default linked-cell structure: subcells have side length >= cut+skin, each subcell is liked to its 3^d - 1 neighbors
    ncell = floor(D/(cut*ALU[2] + skin));
    nbs = 3;
    if (model == 1 || model == 2) ncell = floor(D/(cut*ALU[2] + skin));
    else if (model == 3) ncell = floor(D/(cut*sAA + skin));
// Alternate approach: subcells have side length >= (cut+skin)/2, each subcell is liked to its 5^d - 1 neighbors
// Unforunately this slows code in high d because it produces a ~ (10/3)^d increase in the computational effort to calculate neighindex
    if (cut > 2.49 && d > 3) {
        nbs = 5;
        if (model == 1 || model == 2) ncell = floor(D/((cut*ALU[2] + skin)/2));
        else if (model == 3)  ncell = floor(D/((cut*sAA + skin)/2));
    }
    ncellprev = ncell;  // ncellprev gets used in NPT and/or phi-controlled MD
    cout << "N = " << N << " D = " << D << " ncell = " << ncell << " rho = " << rho << " phi = " << phi << endl;
    
// Set up the linked subcells
    supercell.setupsupercell();
// Set up SWAP Monte Carlo if needed
    if (attfrac) SWAP.setupswap();
// Set up file I/O
    readwrite.openfiles();
// Set up mobility calculations
    diags.initmobile();

// Thread and start the simulation.  The master thread controls things but needsrebuild(),
// buildneighborlist(), leapfrog(), getthermo(), Berendsen(), etc. are all parallelized
    startstep = step;
    thiswindowstart = startstep;
    omp_set_num_threads(nthreads);
#pragma omp master
{
// Initialize neighbor lists
    cout << "Initializing neighbor lists." << endl;
    auto thisnlstart = chrono::high_resolution_clock::now();
    supercell.buildneighborlist();
    auto thisnlfinish = chrono::high_resolution_clock::now();
    nltime = thisnlfinish-thisnlstart;
    totnltime += nltime;
    neighlistbuilds++;
    
// If necessary, execute a "pushoff" to suppress divergent forces associated with particles' random initial positions.
// After pushoff is complete, give particles random initial velocities corresponding to T = Ttarg
// Note the "-12" needs to be adjusted based on model, phi, and d
    if (!readresflag) {
        auto pushstart = chrono::high_resolution_clock::now();
        minims.FIRE(1,0);   // Option 1: partial modified-FIRE minimization
//        minims.steepest(1);  // Option 2: steepest-descent minimization
        for (k = 0; k < d*N; k++) RVF[(3*d)*(k/d) + k%d + d] = 2*sqrt(3)*((1.0*rand())/RAND_MAX - .5)*sqrt(Ttarg/mass[k/d]);
        auto pushfinish = chrono::high_resolution_clock::now();
        pushtime = pushfinish-pushstart;
    }
// Alternatively, if began this run from a restart file, need initialize forces
    if (readresflag) {
        #pragma omp parallel for schedule(guided)
        for (k = 0; k < N; k++) getforce(k);
    }
    
// Start the MD simulation, run for nsteps steps
    cout << "Starting MD run." << endl;
do {
// Set up MSD calculation to measure diffusivity, etc. on step nmsdstart
    if ((!readresflag && (step == 0 || step == nmsdstart)) || (readresflag && step == nmsdstart) || (step >= nmsdstart && !((step-nmsdstart)%nwindow))) {
        Dinit = D;
        for (k = 0; k < d*N; k++) Rinit[k] = RVF[(3*d)*(k/d) + k%d];
        for (k = 0; k < d*N; k++) initimg[k] = image[k];
        if (step == nmsdstart) readwrite.writeRinit(step);
        thiswindowstart = step;
    }
    
// Vary n_thermo to output thermo data more/less often -- useful for getting nice msd, ngp, S(q,t) data
/*    if (step == thiswindowstart) nthermo = 1;     // write data every nbet timesteps for the first tau
    else if (step - thiswindowstart == static_cast<int>(1/dt))  nthermo = static_cast<int>(.04/dt);
    else if (step - thiswindowstart == static_cast<int>(10/dt)) nthermo = static_cast<int>(.2/dt);
    else if (step - thiswindowstart == static_cast<int>(100/dt)) nthermo = static_cast<int>(1/dt);  */

// Check if neighbor list needs to be rebuilt, rebuild it if it does.
    if (supercell.needsrebuild()) {
        auto thisnlstart = chrono::high_resolution_clock::now();
        supercell.buildneighborlist();
        auto thisnlfinish = chrono::high_resolution_clock::now();
        nltime = thisnlfinish-thisnlstart;
        totnltime += nltime;
        neighlistbuilds++;
        avgneigh = 0;
        for (i = 0; i < N; i++) avgneigh += 2.0*numneigh[i]/N;
        totavgneigh += avgneigh;
    }
    
// Calculate and output thermodynamic and other data
    if (step == startstep) cout << "Step t T rho phi P Z Epair namsd ngp f_ov S(q,t) avgmobileclustersize avgimmobileclustersize stringfrac avgstringsize swapsucc swapatt Runtime" << endl;
    if (!(step%nthermo)) {
        auto thisdiagstart = chrono::high_resolution_clock::now();
        diags.getthermo(true);
        totelapsed = thisdiagstart-start;
        cout << totelapsed.count() << endl;
        if (vhcflag && step >= nmsdstart && !((step-nmsdstart)%nvhc)) diags.getvanHove();
        if (gofrflag && step >= nmsdstart && !((step-nmsdstart)%ngofr)) diags.getgofr();
        auto thisdiagfinish = chrono::high_resolution_clock::now();
        diagtime = thisdiagfinish-thisdiagstart;
        totdiagtime += diagtime;
    }
    
// If desired, apply barostat once every nbaro steps: need set Ptarg by editing above code
    if (baroflag && step > 0 && !(step%nbaro)) {
        auto thisbarostart = chrono::high_resolution_clock::now();
        Berendsen(Ptarg);
        auto thisbarofinish = chrono::high_resolution_clock::now();
        barotime = thisbarofinish-thisbarostart;
        totbarotime += barotime;
    }
    
// If desired, write coordinates and/or restart files at regular intervals
    if (!(step%ndump)) readwrite.writecoords(step);
    if (step > 0 && !(step%nrestart)) readwrite.writerestart(step);
    
// If desired, energy-minimize once every nmin timesteps
    if (minflag && step > 0 && step-nmsdstart >= 0 && !((step-nmsdstart)%nmin)) {
        if (step == nmsdstart) readwrite.outputmin(-1,0,0,0);
        auto minstart = chrono::high_resolution_clock::now();
        minims.FIRE(pow(10,-25.0),0);   // FIRE minimization
//        minims.steepest(pow(10,-25.0));  // Option 2: steepest-descent minimization
        supercell.buildneighborlist();
        diags.getthermo(true);       // re-output pre-minimization thermo data
        auto currtime = chrono::high_resolution_clock::now();
        totelapsed = currtime-start;
        cout << totelapsed.count() << endl;
        auto minfinish = chrono::high_resolution_clock::now();
        mintime = minfinish - minstart;
        totmintime += mintime;
    }
    
// If desired, perform SWAP once every nswap steps
    if (attfrac && !(step%nswap)) {
        auto thisswapstart = chrono::high_resolution_clock::now();
        lastswapsucc = swapsucc;
        SWAP.swapmove();
        if (swapsucc > lastswapsucc) supercell.buildneighborlist();
        lastswapstep = step;
        auto thisswapfinish = chrono::high_resolution_clock::now();
        swaptime = thisswapfinish-thisswapstart;
        totswaptime += swaptime;
    }
    
// Here perform time integration: integrate forward in time by nbet timesteps
    auto thisverlstart = chrono::high_resolution_clock::now();
    for (i = 0; i < nbet; i++) leapfrog();
    auto thisverlfinish = chrono::high_resolution_clock::now();
    verltime = thisverlfinish-thisverlstart;
    totverltime += verltime;
    step += nbet;
} while (step < nsteps+1);  // End of the MAIN LOOP
} // End of the "master" pragma

// Dethread and wrap up the simulation
   omp_set_num_threads(1);
   readwrite.closefiles();
   if (attfrac) SWAP.stopswap();
   utils.dealloc();
    
// High resolution timing and diagnostics
   auto finish = chrono::high_resolution_clock::now();
   totelapsed = finish-start;
   cout << neighlistbuilds << " neighbor list builds" << endl;
   cout << "Average # of neighbors = " << totavgneigh/neighlistbuilds << endl;
   cout << "Execution took " << totelapsed.count() << " seconds." <<  endl;
   cout << "Pushoff took " << pushtime.count() << " seconds.  Percentage = " << 100*pushtime.count()/totelapsed.count() << endl;
   cout << "Leapfrog timestepping took " << totverltime.count() << " seconds: Percentage = " << 100*totverltime.count()/totelapsed.count() << endl;
   cout << "Neigh-list builds took " << totnltime.count() << " seconds: Percentage = " << 100*totnltime.count()/totelapsed.count() << endl;
   cout << "Minimizations took " << totmintime.count() << " seconds.  Percentage = " << 100*totmintime.count()/totelapsed.count() << endl;
   cout << "Barostatting took " << totbarotime.count() << " seconds: Percentage = " << 100*totbarotime.count()/totelapsed.count() << endl;
   cout << "SWAPping took " << totswaptime.count() << " seconds: Percentage = " << 100*totswaptime.count()/totelapsed.count() << endl;
   cout << "Thermodynamics calculations/diagnostics took " << totdiagtime.count() << " seconds: Percentage = " << 100*totdiagtime.count()/totelapsed.count() << endl;
   cout << "Other processes took " << totelapsed.count() - totnltime.count() - totverltime.count() - pushtime.count() - totmintime.count() - totbarotime.count() - totswaptime.count() - totdiagtime.count() << " seconds: Percentage = " << 100*(totelapsed.count() - totnltime.count() - totverltime.count() - pushtime.count() - totmintime.count() - totbarotime.count() - totswaptime.count() - totdiagtime.count())/totelapsed.count() << endl;
   for (i = 0; i < 3; i++) cout << endl;
    
   return 0;
}



// Leapfrog integrator
// Includes a Langevin thermostat and a Berendsen thermostat for NVT runs
void leapfrog() {
    int i, k;
    double K = 0;  // total kinetic energy of system
    double Ktarg = d*N*Ttarg/2;
    double dampstepT = tau_temp/dt;     // Berendsen thermostat damping time
    double gam = 1.0/tau_temp;  // Langevin damping-force prefactor
    double gfac = sqrt(24*gam*Ttarg/dt);  // Langevin thermal-noise-force prefactor
    double psum[d];     // for momentum-zeroing
#pragma omp parallel
    {
    unsigned int rseed = rand() + omp_get_thread_num();
// First half of leapfrog; update v and x, map atoms back into box, reset all forces to zero
    #pragma omp for schedule(static,nchunk) reduction(+:K)
    for (k = 0; k < d*N; k++) {
        RVF[(3*d)*(k/d) + k%d + d] += dthalf*RVF[(3*d)*(k/d) + k%d + 2*d]/mass[k/d];
        RVF[(3*d)*(k/d) + k%d] += dt*RVF[(3*d)*(k/d) + k%d + d];
        if (RVF[(3*d)*(k/d) + k%d] > hi) {  RVF[(3*d)*(k/d) + k%d] -= D;   image[k] += 1;  }
        else if (RVF[(3*d)*(k/d) + k%d] < lo) {  RVF[(3*d)*(k/d) + k%d] += D;   image[k] -= 1;  }
        RVF[(3*d)*(k/d) + k%d + 2*d] = 0;
// Uncomment if using Berendsen thermostat
        if (!baroflag && !(step%nbaro)) K += mass[k/d]*RVF[(3*d)*(k/d) + k%d + d]*RVF[(3*d)*(k/d) + k%d + d]/2;
    }
// Uncomment if using Berendsen thermostat
    double lambda;  // Berendsen velocity-scaling factor
    if (!baroflag && !(step%nbaro)) lambda = sqrt(1 + nbaro*(Ktarg/K - 1)/dampstepT);
        
// Update forces
    #pragma omp for schedule(guided)
    for (k = 0; k < N; k++) getforce(k);
        
// Second half of leapfrog: update v
// Apply Berendsen velocity-rescaling factor ONLY if NOT also employing Berendsen baro/thermostat
    #pragma omp for schedule(static,nchunk) 
    for (k = 0; k < d*N; k++) {
// Uncomment the below two lines if using Berendsen thermostat
        RVF[(3*d)*(k/d) + k%d + d] += dthalf*RVF[(3*d)*(k/d) + k%d + 2*d]/mass[k/d];
        if (!baroflag  && !(step%nbaro)) RVF[(3*d)*(k/d) + k%d + d] *= lambda;
// Uncomment the below two lines if using Langevin thermostat
//        RVF[(3*d)*(k/d) + k%d + 2*d] += gfac*((1.0*rand_r(&rseed))/RAND_MAX - 0.5) - gam*RVF[(3*d)*(k/d) + k%d + d];
//        RVF[(3*d)*(k/d) + k%d + d] += dthalf*RVF[(3*d)*(k/d) + k%d + 2*d]/mass[k/d];
    }
        
// Zero out the total system momentum
// Only really need this for very sensitive measurements of msd (e.g. for glass-transition studies)
// Otherwise can comment it out (since this is also done every time NLs get built)
    #pragma omp for schedule(static,nchunk) reduction(+:psum)
    for (k = 0; k < d*N; k++) psum[k%d] += mass[k/d]*RVF[(3*d)*(k/d) + d + k%d];
    #pragma omp for schedule(static,nchunk)
    for (k = 0; k < d*N; k++) RVF[(3*d)*(k/d) + d + k%d] -= psum[k%d]/(N*mass[k/d]);
    } // end of the parallel region
    
    return;
}



// Berendsen barostat
// This version implicitly assumes the isothermal compressibility kappa = 1
void Berendsen(double Ptarg) {
// Call getthermo() if it wasn't already called this timestep
    if (step%nthermo) diags.getthermo(false);
    rho = N/vol;
    double Pdamp = 18.257; // Berendsen barostat pressure damping time
    double Tdamp = 5.7735; // Berendsen thermostat temperature damping time
    double dampstepP = tau_press/dt;
    double dampstepT = tau_temp/dt;
    double deltaP = P - Ptarg;
    double deltaT = T - Ttarg;
    double lambda, Vexpfac, Lexpfac; // Berensen velocity, volume, and length scaling factors
    int k;
#pragma omp master
{
    Vexpfac = 1 + nbaro*deltaP/dampstepP;
    Lexpfac = pow(Vexpfac,1.0/d);
    lambda = 1 - nbaro*(deltaT/T)/dampstepT;
// Adjust V based on sign and magniture of delta P, # of steps between barostat calls
    vol *= Vexpfac;
    D *= Lexpfac;   // adjust side length of cubic cells
    lo = -D/2;  hi = D/2;
}
// Parallelize the position, velocity updates
    #pragma omp parallel
{
// Scale all positions by Lexpfac and velocities, map back into box if necessary
// Then scale all velocities by lambda/Lexpfac
    #pragma omp for schedule(static,nchunk)
    for (k = 0; k < d*N; k++) {
        RVF[(3*d)*(k/d) + k%d] *= Lexpfac;
        RVF[(3*d)*(k/d) + k%d + d] *= lambda/Lexpfac;
    }
}
// Adjust number of cells if needed
    int ncelltmp = floor(D/(cut*ALU[2] + skin));
    if (ncelltmp != ncellprev) {
        ncell = ncelltmp;
        supercell.setupsupercell();
        supercell.buildneighborlist();
        neighlistbuilds++;
        ncellprev = ncell;
    }
    else if (supercell.needsrebuild()) {
        supercell.buildneighborlist();
        neighlistbuilds++;
    }
    
    return;
}


// Calculate force on atom i using new generalized approach valid for arbitrary force laws
void getforce(int i) {
    int j, k, l, u;
    double f[d]; // force on atom i due to atom u
    double fovr;  // this force's magnitude/del
    double delc[d],delc0[d],delsq, del;
    double epsilonij, sigmaij, sigmaijsq; // For polydisp force-mixing
    double facij;
    
// Calculate force f on atom i by looping over its neighbors (j)
    for (k = 0; k < d; k++) f[k] = 0;
// Defining delc0 here reduces # of sequential accesses of distant parts of memory
    for (k = 0; k < d; k++) delc0[k] = -RVF[3*d*i+k];
// Loop over atom i's neighbors
    for (j = 0; j < numneigh[i]; j++) {
        u = allneigh[i*maxneigh + j];
// Apply mixing rules for particle diameters, force/energy prefactors
        sigmaij = forces.getsigmaij(i,u);
        epsilonij = forces.getepsilonij(i,u);
        sigmaijsq = sigmaij*sigmaij;
        delsq = 0;
        #pragma ivdep
        for (k = 0; k < d; k++) {
            delc[k] = delc0[k] + RVF[3*d*u+k];
            if (delc[k] < -D/2.0) delc[k] += D;    else if (delc[k] > D/2.0) delc[k] -= D;
            delsq += delc[k]*delc[k];
            if (d > 5) if (delsq > cutsq*sigmaijsq) goto nextneigh;
        }
// Get |F|/r and update the forces on the interacting particles
        if (d > 5 || delsq < cutsq*sigmaijsq) {
            fovr = forces.getfovr(delsq,sigmaij,sigmaijsq,epsilonij);
            #pragma ivdep
            for (k = 0; k < d; k++) {
                f[k] = -delc[k]*fovr;
                RVF[3*d*i+2*d+k] += f[k];    RVF[3*d*u+2*d+k] -= f[k];
            }
        }
        nextneigh: ;
    }
    return;
}




